package org.codehaus.xfire.aegis.type;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ognl.Ognl;
import ognl.OgnlException;

import org.codehaus.xfire.SOAPConstants;
import org.codehaus.xfire.aegis.AegisService;
import org.codehaus.xfire.aegis.mapping.TypeRegistry;
import org.codehaus.xfire.fault.XFireFault;
import org.dom4j.Attribute;
import org.dom4j.Element;
import org.dom4j.Namespace;
import org.dom4j.QName;

/**
 * <p>
 * The BeanType takes child types and read/writes each of them.
 * </p>
 * <p>
 * The BeanType can take its own OGNL expression and a key.  This is for when
 * you want to take action when <i>reading</i> in an <code>Element</code>.
 * For instance, if you specify the OGNL expression as
 * "new com.company.Bean()" and the key as "bean" it will put a
 * new Bean object in the context with the key "bean".
 * </p>
 * 
 * @author <a href="mailto:dan@envoisolutions.com">Dan Diephouse</a>
 */
public class BeanType
	extends Type
{
    private List children;

    private String key;
    
    public void write( Element element, Map context )
        throws XFireFault
    {
        if ( getKey() == null || context.containsKey(getKey()) )
        {
            Element typeEl = element.addElement( getQName() );
    
            if ( getChildren() != null )
            {
                for ( Iterator itr = getChildren().iterator(); itr.hasNext(); )
                {
                    Type type = (Type) itr.next();
                    type.write(typeEl, context);
                }
            }
        }
    }
    
    public void read( Element element, Map context )
        throws XFireFault 
    {
        read( element, 0, context );
    }

    /**
     * @see org.codehaus.xfire.aegis.type.Type#read(org.dom4j.Element, int, java.util.Map)
     */
    public void read(Element element, int occurrence, Map context) throws XFireFault
    {
        List elements = element.elements( getQName() );
        if ( elements == null || elements.size() <= occurrence )
            return;
        
        Element typeEl = (Element) elements.get(occurrence);
        
        if ( getOgnl() != null )
        {
			try
			{
				Object value = Ognl.getValue( getOgnl(), context, (Object) null );
                
                context.put( key, value );
			}
			catch (OgnlException e)
			{
				throw new XFireFault( "Couldn't process message.", e, XFireFault.SENDER );
			}
        }
        
        if ( getChildren() != null )
        {
            for ( Iterator itr = getChildren().iterator(); itr.hasNext(); )
            {
                Type type = (Type) itr.next();
                type.read(typeEl, context);
            }
        }
    }

    public List getChildren()
    {
        return children;
    }
    
    public void setChildren( List children )
    {
        this.children = children;
    }
    
    public void addChild( Type type )
    {
        if ( children == null )
            children = new ArrayList();
        
        children.add( type );
    }    
    
	public String getKey()
	{
		return key;
	}
    
	public void setKey(String key)
	{
		this.key = key;
	}

	public boolean isComplex()
	{
		return true;
	}
    
	/**
	 * @see org.codehaus.xfire.aegis.type.Type#writeSchema(org.dom4j.Element)
	 */
	public void writeSchema(Element root)
	{
        Namespace xsdNs = root.getNamespaceForURI( SOAPConstants.XSD );
        Namespace beanNs = root.getNamespaceForURI( getSchemaType().getNamespaceURI() );
        
        org.dom4j.QName elementQ = new org.dom4j.QName( "element", xsdNs );
        Element concreteEl = root.addElement( elementQ );
        concreteEl.addAttribute("name", getName());
        //concreteEl.addAttribute("type", beanNs.getPrefix() + ":" + getSchemaType().getName());
        
        org.dom4j.QName complexQ = new org.dom4j.QName("complexType", xsdNs); 
        Element complex = concreteEl.addElement( complexQ );
        // complex.addAttribute( "name", getName() );

        org.dom4j.QName seqQ = new org.dom4j.QName("sequence", xsdNs); 
        Element seq = complex.addElement( seqQ );

        if ( getChildren() == null )
            return;
        
        for ( Iterator itr = getChildren().iterator(); itr.hasNext(); )
        {
            Type type = (Type) itr.next();

            Element element = seq.addElement( elementQ );
            
            Namespace typeNS = root.getNamespaceForURI( type.getSchemaType().getNamespaceURI() );
            if ( type.isComplex() )
            {
                 element.addAttribute( "ref",  typeNS.getPrefix() + ":" + type.getSchemaType().getName() );
            }
            else
            {
                element.addAttribute( "name", type.getName() );
                // TODO: Add config support for nillable
                element.addAttribute( "nillable", "true" );
                element.addAttribute("type", typeNS.getPrefix() + ":" + type.getSchemaType().getName());
            }
        }
	}

	/**
	 * @see org.codehaus.xfire.aegis.type.Type#configure(org.dom4j.Element)
	 */
	public void configure(Element configuration, AegisService service, TypeRegistry reg )
	{
		setName( configuration.attributeValue("name") );
        setOgnl( configuration.attributeValue("ognl") );
        setKey( configuration.attributeValue("key") );
        
        String min = configuration.attributeValue("minOccurs");
        if (min != null)
        	setMinOccurs(min);
        
        String max = configuration.attributeValue("maxOccurs");
        if (max != null)
        	setMaxOccurs(max);
        
        setQName( QName.get( getName(), service.getDefaultNamespace() ) );
        setDocumentation( configuration.getTextTrim() );
        
        Attribute xsdAt = configuration.attribute("schemaType");
        if ( xsdAt != null )
        {
            setSchemaType( null );
        }
        else
        {
            setSchemaType( QName.get(getName(), service.getDefaultNamespace() ) );
        }
        
        List childElements = configuration.elements();
        for ( Iterator itr = childElements.iterator(); itr.hasNext(); )
        {
            Element paramEl = (Element) itr.next();
            QName typeQ = QName.get(paramEl.getName(), service.getSoapVersion());

            Type type = reg.createType(typeQ);
            
            if ( type == null )
                throw new RuntimeException("No such type: " + typeQ.getName());
            
            type.configure(paramEl, service, reg);
            
            addChild(type);
        }
    }

	public Set getDependencies()
	{
        if ( children == null )
            return null;
        
		Set deps = new HashSet();
        
        deps.addAll( children );
        
        return deps;
	}

}
